;; Generated package description from openpgp.el  -*- no-byte-compile: t -*-
(define-package "openpgp" "1.0.2" "Client for keys.openpgp.org" '((emacs "24.4")) :commit "69254001eae10b3b162200082bcbc684363e84d7" :authors '(("Philip Kaludercic" . "philipk@posteo.net")) :maintainer '("Philip Kaludercic" . "~pkal/public-inbox@lists.sr.ht") :keywords '("convenience") :url "https://git.sr.ht/~pkal/openpgp/")
